class NoTasksLoadedError(Exception):
    pass


class ExtensionsDirectoryNotFoundError(Exception):
    pass


class NotAPackageError(Exception):
    pass


class CheckpointInvalidError(Exception):
    pass
